// E-flio B - Introduo  Programao 2024/25

#include <stdlib.h>
#include <stdio.h>

///////////////////////////////////////////////////////////////////////////////////////////////
// Tipos: TAbelha, TCelula, TFavo, TColmeia

// nmero mximo de clulas num favo
#define CELULAS_FAVO 3000
// dias de idade mxima de uma abelha operria
#define IDADE_MAXIMA 40
// miligramas de polen por clula (e montante carregado por uma abelha)
#define POLEN_CELULA 20
// miligramas de nectar por clula (e montante carregado por uma abelha)
#define NECTAR_CELULA 40
// miligramas de mel por clula 
#define MEL_CELULA 500
// consumo de polen e mel por dia (miligramas)
#define POLEN_ABELHA 4
#define MEL_ABELHA 10
#define TEMPO_DIA 1440

///////////////////////////////////////////////////////////////////////////////////////////////
// TAbelha: tipo que guarda informao sobre cada abelha 

// enum funcaoAbelha: tipo enumerado com a lista de funes das abelhas, por ordem do enunciado
typedef enum {
	faxineira = 0, nutriz, construtora, guardia, forrageira,
	zangao, rainha, ovo, larva, pupa, semFuncao
} TFuncaoAbelha;

typedef struct SAbelha {
	int nascimento;		
	int ultimaRefeicao; 
	TFuncaoAbelha funcao;
	struct SAbelha* prox;
} TAbelha;

// NovaAbelha: adiciona uma nova abelha a uma lista, retornando a lista atualizada
TAbelha* NovaAbelha(TFuncaoAbelha funcao, int nascimento, int ultimaRefeicao, TAbelha* lista) {
	TAbelha* abelha;
	abelha = (TAbelha*)malloc(sizeof(TAbelha));
	if (abelha != NULL) {
		abelha->nascimento = nascimento;
		abelha->funcao = funcao;
		abelha->ultimaRefeicao = ultimaRefeicao;
		abelha->prox = lista;
		lista = abelha;
	}
	return lista;
}

// ContarAbelhas: conta o nmero de abelhas numa lista, que tem uma dada funo
int ContarAbelhas(TFuncaoAbelha funcao, TAbelha* lista) {
	int contagem = 0;
	while (lista != NULL) {
		if (lista->funcao == funcao)
			contagem++;
		lista = lista->prox;
	}
	return contagem;
}

// AntigaAbelha: retorna a abelha mais antiga da lista, numa dada funo
TAbelha* AntigaAbelha(TFuncaoAbelha funcao, TAbelha* lista) {
	TAbelha* abelha = NULL;
	while (lista != NULL) {
		if (lista->funcao == funcao &&
			(abelha == NULL || abelha->nascimento > lista->nascimento))
			abelha = lista;
		lista = lista->prox;
	}
	return abelha;
}

// CriarAbelhas: cria uma nova lista de abelhas, de acordo com o enunciado
TAbelha* CriarAbelhas(int nAbelhas) {
	TAbelha* lista = NULL;
	int i, j;
	if (nAbelhas < IDADE_MAXIMA*5) // tem de existir abelhas de todos os tipos
		return NULL;
	lista = NovaAbelha(rainha, 0, 0, NULL); // comear pela rainha
	for (i = 1; i <= IDADE_MAXIMA; i++) { // processar todas as idades
		lista = NovaAbelha(zangao, -i * TEMPO_DIA, -i*TEMPO_DIA/IDADE_MAXIMA, lista); // em cada dia h um zango 
		for (j = 0; j < nAbelhas / IDADE_MAXIMA; j++) // h N/40 operrias em cada dia
			lista = NovaAbelha(faxineira + j % 5, // funo, vai percorrendo a lista, utilizando a ordem do tipo enumerado
				-i * TEMPO_DIA + j, // nascimento no dia certo, somando 1 minuto em cada abelha
				-i * TEMPO_DIA / IDADE_MAXIMA,  // instante da ltima refeio
				lista);
	}
	return lista;
}

// LibertarAbelha: apaga uma abelha especfica da lista, retornando a lista atualizada
TAbelha* LibertarAbelha(TAbelha* abelha, TAbelha* lista) {
	TAbelha* retornar = lista;
	if (lista == NULL || abelha == NULL)
		return retornar;
	if (abelha == lista) { 
		retornar = lista->prox;
		free(abelha);
		return retornar;
	}
	while (lista!=NULL && lista->prox != NULL) {
		if (lista->prox == abelha) {
			lista->prox = abelha->prox;
			free(abelha);
			return retornar;
		}
		lista = lista->prox;
	}
	return retornar;
}

// LibertarAbelhas: liberta toda a lista
void LibertarAbelhas(TAbelha* lista) {
	TAbelha* prox;
	while (lista != NULL) {
		prox = lista->prox;
		free(lista);
		lista = prox;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
// TCelula: tipo que guarda informao sobre uma clula
// union: como o contedo depende do tipo de clula, optou-se pelo o union para o "conteudo"

// enum tipoCelula: tipos de clulas
typedef enum { mel = 0, polen, nectar, cria, criaZangao, realeira } TTipoCelula;

typedef struct SCelula {
	TTipoCelula tipo;
	union {
		TAbelha *abelha; // ovo, larva, pupa (cria ou zango)
		int peso; // quantidade de mel, polen ou nectar, em miligramas
	} conteudo;
	struct SCelula* prox;
} TCelula;

// NovaCelula: adiciona uma nova clula  lista, retornando a lista atualizada
TCelula* NovaCelula(TTipoCelula tipo, TCelula* lista) {
	static int quantidade[] = { MEL_CELULA, POLEN_CELULA, NECTAR_CELULA };
	TCelula* celula;
	celula = (TCelula*)malloc(sizeof(TCelula));
	if (celula != NULL) {
		celula->tipo = tipo;
		// inicialmente vazia a clula de cria ou superior
		if (celula->tipo >= cria)
			celula->conteudo.abelha = NULL;
		else 
			// clula para recursos iniciam-se cheias
			celula->conteudo.peso = quantidade[tipo];
		celula->prox = lista;
		lista = celula;
	}
	return lista;
}

// ContarCelulas: retorna o nmero de clulas de cada tipo e estado (vazia ou usada)
int ContarCelulas(TTipoCelula tipo, int vazia, TCelula* lista) {
	int contagem = 0;
	int estaVazio;
	while (lista != NULL) {
		if (lista->tipo == tipo) {
			estaVazio = (lista->tipo >= cria ?
				lista->conteudo.abelha == NULL :
				lista->conteudo.peso == 0);
			if (vazia?estaVazio:!estaVazio) 
				contagem++;
		}
		lista = lista->prox;
	}
	return contagem;
}


// CriarCelulas: cria uma lista de clulas
TCelula* CriarCelulas(int nCelulas) {
	TCelula* lista = NULL;
	int i, j;
	if(nCelulas>0)
		lista = NovaCelula(criaZangao, NULL); // uma clula para zanges
	for (i = 0; i < nCelulas; i++) // o resto para dividir por igual
		lista = NovaCelula(mel + i % 4, lista);
	return lista;
}

// LibertarCelulas: liberta todas as clulas da lista
void LibertarCelulas(TCelula* lista) {
	TCelula* prox;
	while (lista != NULL) {
		prox = lista->prox;
		free(lista);
		lista = prox;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
// TFavo: tipo que guarda informao sobre um favo
typedef struct SFavo {
	int criacao; // data de criao
	TCelula *celulas;
	struct SFavo* prox;
} TFavo;

// NovoFavo: cria um novo favo e clulas respetivas
TFavo* NovoFavo(int nCelulas, int criacao, TFavo* lista) {
	TFavo* favo;
	favo = (TFavo*)malloc(sizeof(TFavo));
	if (favo != NULL) {
		favo->celulas = CriarCelulas(nCelulas);
		favo->criacao = criacao;
		favo->prox = lista;
		lista = favo;
	}
	return lista;
}

// CriarFavos: cria um conjunto de favos, a infraestrutura da colmeia
TFavo* CriarFavos(int nCelulas) {
	TFavo* lista = NULL;
	int i, j, nFavos;
	nFavos = 1 + (nCelulas - 1) / CELULAS_FAVO;
	for (i = 0; i < nFavos; i++)
		lista = NovoFavo(nCelulas / nFavos, i, lista);
	return lista;
}

// FavoMaisNovo: retorna o favo com a maior data de criao
int FavoMaisNovo(TFavo* lista) {
	int criacao = 0;
	while (lista != NULL) {
		if (lista->criacao > criacao)
			criacao = lista->criacao;
		lista = lista->prox;
	}
	return criacao;
}

// LibertarFavos: liberta todos os favos e clulas
void LibertarFavos(TFavo* lista) {
	TFavo* prox;
	while (lista != NULL) {
		prox = lista->prox;
		LibertarCelulas(lista->celulas);
		free(lista);
		lista = prox;
	}
}


///////////////////////////////////////////////////////////////////////////////////////////////
// TColmeia: tipo que guarda informao sobre toda a colmeia
typedef struct {
	int tempo; // tempo em minutos, sendo 0 o instante de criao
	// gesto da colmeia
	int naoTrabalham; // 1 apenas h a alimentao, mas no h promoo nem trabalho
	double racioFuncao[5]; // percentagem de operrias por funo
	int prontaTrabalhar[5]; // operrias prontas para trabalhar (acumula com os minutos anteriores)
	TFavo *favos; // infraestrutura da colmeia
	TAbelha* abelhas; // recursos "humanos/abelhas" da colmeia
} TColmeia;

// CriarColmeia: criao de uma colmeia, com nmero de abelhas e clulas
TColmeia * CriarColmeia(int nAbelhas, int nCelulas) {
	static double racioFuncao[] = { 0.1, 0.3, 0.2, 0.1, 0.3 }; // inicializar o rcios fixos
	TColmeia* colmeia;
	int i;
	colmeia = (TColmeia*)malloc(sizeof(TColmeia));
	if (colmeia != NULL) {
		colmeia->tempo = 0;
		colmeia->naoTrabalham = 0;
		for (i = 0; i < 5; i++) {
			colmeia->racioFuncao[i] = racioFuncao[i];
			colmeia->prontaTrabalhar[i] = 0;
		}
		colmeia->abelhas = CriarAbelhas(nAbelhas);
		colmeia->favos = CriarFavos(nCelulas);
	}
	return colmeia;
}

// CelulaColmeia: 
// Caso seja uma clula de recursos e abelha!=NULL, retornar a clula em consumo, ou seja, 
// a clula com recursos, mas com menos recursos 
// Caso abelha seja NULL retornar uma clula desse tipo vazia
// Se for uma clula de cria, e abelha for NULL retornar uma clula vazia 
// Se abelha!=NULL, procurar pela clula com essa abelha para limpar o contedo da clula
TCelula* CelulaColmeia(TTipoCelula tipo, TAbelha *abelha, TColmeia* colmeia) {
	TCelula* emConsumo=NULL, *celula;
	TFavo* favo = colmeia->favos;
	while (favo != NULL) {
		celula = favo->celulas;
		while (celula != NULL) {
			if (tipo == cria) {
				if (abelha != NULL) {
					if (celula->tipo == cria && celula->conteudo.abelha == abelha) {
						celula->conteudo.abelha = NULL;
						return NULL;
					}
				} else if (celula->tipo == cria && celula->conteudo.abelha == NULL)
					return celula;
			}
			else if (abelha == NULL) { // apenas uma clula vazia
				if (celula->tipo == tipo &&
					celula->conteudo.peso==0)
					return celula;
			} else if (celula->tipo == tipo &&
				celula->conteudo.peso > 0 &&
				(emConsumo == NULL || 
					emConsumo->conteudo.peso > celula->conteudo.peso))
				emConsumo = celula;
			celula = celula->prox;
		}
		favo = favo->prox;
	}
	return emConsumo;
}

// AlimentarColmeia: ver que abelhas precisam de refeio, e se tem excesso de idade
void AlimentarColmeia(TColmeia* colmeia) {
	TAbelha* abelha = colmeia->abelhas, *aux;
	TCelula* celMel = NULL, * celPolen = NULL;
	// cada abelha precisa de Polen e Mel por dia
	while (abelha != NULL) {
		if (abelha->funcao > rainha) {
			// ovo, larva ou pupa, no  adulta, tem alimentao distinta 
			// (a alimentao da larva no  contemplada no e-flio)
			abelha = abelha->prox;
		} else if (colmeia->tempo - abelha->nascimento >= TEMPO_DIA * IDADE_MAXIMA &&
			abelha->funcao!=rainha) { // a rainha vive 1 a 2 anos
			// abelha morre de idade
			aux = abelha->prox;
			colmeia->abelhas = LibertarAbelha(abelha, colmeia->abelhas);
			abelha = aux;
		} else if (colmeia->tempo - abelha->ultimaRefeicao >= TEMPO_DIA) {
			// alimentar abelha ou morre de fome
			if(celMel==NULL || celMel->conteudo.peso==0) 
				celMel = CelulaColmeia(mel, abelha, colmeia);
			if (celPolen == NULL || celPolen->conteudo.peso == 0)
				celPolen = CelulaColmeia(polen, abelha, colmeia);
			if (celMel != NULL && celPolen != NULL) {
				celMel->conteudo.peso -= MEL_ABELHA;
				celPolen->conteudo.peso -= POLEN_ABELHA;
				abelha->ultimaRefeicao = colmeia->tempo;
				abelha = abelha->prox;
			}
			else {
				// abelha morre de fome
				aux = abelha->prox;
				colmeia->abelhas = LibertarAbelha(abelha, colmeia->abelhas);
				abelha = aux;
			}
		}
		else
			abelha = abelha->prox;
	}
}

// PromoverColmeia: ver se h abelhas para promover
void PromoverColmeia(TColmeia* colmeia) {
	static int tempoPromocao[] = { 3,9,21 };
	TAbelha* abelha;
	TFuncaoAbelha funcao = forrageira;
	int nAbelhas[5]; // nmero de abelhas de cada funo (operrias)
	int totalAbelhas = 0;
	// contar abelhas de cada funo
	for (funcao = faxineira; funcao <= forrageira; funcao++) {
		nAbelhas[funcao] = ContarAbelhas(funcao, colmeia->abelhas);
		totalAbelhas += nAbelhas[funcao];
	}
	// promover se tiver a baixo do valor, promover uma abelha da funo inferior
	for (funcao = forrageira; funcao > faxineira; funcao--) {
		if (nAbelhas[funcao] < colmeia->racioFuncao[funcao] * totalAbelhas &&
			nAbelhas[funcao - 1]>0) {
			abelha = AntigaAbelha(funcao - 1, colmeia->abelhas);
			if(abelha!=NULL)
				abelha->funcao++;
		}
	}
	// promover as abelhas no adultas
	abelha = colmeia->abelhas;
	while (abelha != NULL) {
		if (abelha->funcao >= ovo && abelha->funcao <= pupa) {
			if (colmeia->tempo - abelha->nascimento >= tempoPromocao[abelha->funcao - ovo] * TEMPO_DIA) 
				abelha->funcao++; // passa de ovo > larva > pupa > semFuncao
			if (abelha->funcao == semFuncao) { // abelha ficou adulta
				// limpar a clula onde a abelha estava
				CelulaColmeia(cria, abelha, colmeia);
				abelha->funcao=faxineira; // passa a operria, primeiro escalo
				abelha->nascimento = colmeia->tempo; // nascimento adulto
			}
		}
		abelha = abelha->prox;
	}
}

// PorOvosColmeia: colocao de um ovo numa clula de cria
void PorOvosColmeia(TColmeia* colmeia) {
	// necessrio 10 faxineiras para limpar, e nesse caso a rainha coloca um ovo
	TCelula* celula = CelulaColmeia(cria, NULL, colmeia);
	if (celula!=NULL && colmeia->prontaTrabalhar[faxineira] >= 10 &&
		ContarAbelhas(rainha,colmeia->abelhas)>0) { // tem de ainda existir a rainha
		colmeia->prontaTrabalhar[faxineira] -= 10;
		celula->conteudo.abelha = colmeia->abelhas = 
			NovaAbelha(ovo, colmeia->tempo, colmeia->tempo, colmeia->abelhas);
	}
}

// ProcurarRecursosColmeia: procurar por polen ou nectar
void ProcurarRecursosColmeia(TColmeia *colmeia) {
	TCelula* celula;
	int procNectar = 0; // procura sempre por plen, 
	// necessrio 60 minutos para uma abelha ir buscar polen ou nectar, considerar 60 abelhas.minuto
	// a clula tem tambm de ser limpa
	celula = CelulaColmeia(procNectar ? nectar : polen, NULL, colmeia);
	if (celula != NULL && colmeia->prontaTrabalhar[forrageira] >= 60 ) {
		colmeia->prontaTrabalhar[forrageira] -= 60;
		colmeia->prontaTrabalhar[faxineira] -= 10; // se faxineiras no estiverem disponveis, a rainha no coloca ovos
		celula->conteudo.peso = (procNectar ? NECTAR_CELULA : POLEN_CELULA);
	}
	else if (celula == NULL)
		colmeia->prontaTrabalhar[forrageira] = 0; // quantidade de trabalho desperdiada
}


// TrabalharColmeia: atualizar quantidade de trabalho de cada tipo disponvel, e 
// - rainha pode num minuto, colocar um novo ovo, se existir uma clula de cria ou criaZangao limpa
// - nutrizes, alimentar as larvas (do dia 3 ao dia 9), passando depois a pulpa --- cada alimentao 1 minuto,
//   tendo de passar 15 vezes, em cada vez o consumo dirio de uma abelha adulta --- larvas no alimentadas, morrem
// - construtoras, se existirem 10 ou mais livres, iniciar construo de uma nova clula, no favo mais vazio
//   procurar respeitar a precentagem ideal de tipos de clulas. 
// - guardians, se existir perigo, atacar
// - forrageiras, procurar nectar/polen na proporo ideal (depende das que esto em transito)
// - em cada viagem, o tempo de viagem  entre 30 e 60 minutos
void TrabalharColmeia(TColmeia* colmeia) {
	TAbelha* abelha = colmeia->abelhas;
	TFuncaoAbelha funcao;
	while (abelha != NULL) {
		// se a abelha operria no est a comer, coloc-la pronta a trabalhar
		if (abelha->funcao <= forrageira && abelha->ultimaRefeicao != colmeia->tempo)
			colmeia->prontaTrabalhar[abelha->funcao]++;
		abelha = abelha->prox;
	}
	// operaes
	PorOvosColmeia(colmeia);
	//	AlimentarLarvasColmeia(colmeia); // (nut) fora do e-flio 
	//	ConstruirColmeia(colmeia); // (con) fora do e-flio
	ProcurarRecursosColmeia(colmeia);
	// .... (gua), nectar>mel...
}

// LibertarColmeia: liberta todas as estruturas de dados
void LibertarColmeia(TColmeia* colmeia) {
	LibertarAbelhas(colmeia->abelhas);
	LibertarFavos(colmeia->favos);
}

///////////////////////////////////////////////////////////////////////////
// Relatrio

// MostraTempo: funo para mostrar os minutos agregados em horas e dias
void MostraTempo(int minutos) {
	int dias, horas;
	dias = minutos / TEMPO_DIA;
	minutos -= dias * TEMPO_DIA;
	horas = minutos / 60;
	minutos -= horas * 60;
	if (dias > 0)
		printf("%d dia(s) ", dias);
	if (horas > 0)
		printf("%d hora(s) ", horas);
	if (minutos > 0)
		printf("%d minuto(s)", minutos);
}

void MostrarQuantidade(int valor) {
	if (valor > 0)
		printf(" %3d", valor);
	else
		printf("    ");
}

// RelatorioCelulas: mostra a parte do relatrio de favos, das suas clulas
void RelatorioCelulas(TCelula* lista) {
	printf("            celulas vazias:");
	for (TTipoCelula tipo = mel; tipo <= realeira; tipo++)
		MostrarQuantidade(ContarCelulas(tipo, 1, lista));
	printf("\n                                 usadas:");
	for (TTipoCelula tipo = mel; tipo <= realeira; tipo++)
		MostrarQuantidade(ContarCelulas(tipo, 0, lista));
}

// RelatorioFavos: mostra o relatrio do favo com determinada data de criao
void RelatorioFavos(TFavo* lista, int criacao) {
	while (lista != NULL) {
		if (lista->criacao == criacao) {
			printf("\n    Favo %3d:", lista->criacao);
			RelatorioCelulas(lista->celulas);
		}
		lista = lista->prox;
	}
}

// RelatorioColmeia: mostra o relatrio da colmeia, o nico solicitado no e-flio
void RelatorioColmeia(TColmeia* colmeia) {
	int criacao, maisNovo;
	printf("\nRelatorio colmeia: ");
	MostraTempo(colmeia->tempo);
	printf("\n fax nut con gua for zan rai ovo lar pup mel pol nec cri zan rea \n");
	for (TFuncaoAbelha funcao = faxineira; funcao < semFuncao; funcao++)
		MostrarQuantidade(ContarAbelhas(funcao, colmeia->abelhas));
	maisNovo = FavoMaisNovo(colmeia->favos);
	for (criacao = 0; criacao <= maisNovo; criacao++)
		RelatorioFavos(colmeia->favos, criacao);
}

///////////////////////////////////////////////////////////////////////////
// Simulao

// SimularColmeia: avana 1 minuto simulando as operaes na colmeia
void SimularColmeia(TColmeia* colmeia) {
	colmeia->tempo++; // avana o tempo

	// - alimentar todas as abelhas que esteja na colmeia, se j passou 24h desde a ultima refeio
	// - caso uma abelha esteja em excesso de idade, remover em vez de alimentar
	AlimentarColmeia(colmeia);

	if (colmeia->naoTrabalham == 0) { // alnea D
		// - promover abelhas por idade, procurando respeitar a proporo pretendida entre funes
		PromoverColmeia(colmeia);

		// coloca as abelhas a trabalhar, conforme a funo
		TrabalharColmeia(colmeia);
	}
}

// TemporadaColmeia: simula uma temporada, parando caso no existam abelhas
void TemporadaColmeia(int tempo, TColmeia* colmeia) {
	while (tempo > 0 && colmeia->abelhas != NULL) {
		tempo--;
		SimularColmeia(colmeia);
	}
}

// FilmeColmeia: sequncia de temporadas intercaladas com um relatrio de colmeia
void FilmeColmeia(int tempo, int temporadas, TColmeia* colmeia) {
	RelatorioColmeia(colmeia);
	while (--temporadas > 0 && colmeia->abelhas != NULL) {
		TemporadaColmeia(tempo, colmeia);
		RelatorioColmeia(colmeia);
	}
}

///////////////////////////////////////////////////////////////////////////
// main: execuo das 4 alneas
int main() 
{
	int nAbelhas, nCelulas, minutos, temporadas;
	TColmeia *colmeia;

	//printf("Quantidade de abelhas: ");
	scanf("%d", &nAbelhas);
	//printf("Quantidade de celulas: ");
	scanf("%d", &nCelulas);
	//printf("Minutos por temporada: ");
	scanf("%d", &minutos);
	//printf("Temporadas: ");
	scanf("%d", &temporadas);
	colmeia = CriarColmeia(nAbelhas, nCelulas);
	if (colmeia != NULL) {
		if (temporadas < 0) { // alnea C
			temporadas *= -1;
			colmeia->naoTrabalham = 1;
		}
		FilmeColmeia(minutos, temporadas + 1, colmeia);
		LibertarColmeia(colmeia);
	}
}
