// E-flio B - Introduo  Programao 2024/25

#include <stdlib.h>
#include <stdio.h>

///////////////////////////////////////////////////////////////////////////////////////////////
// Tipos: TAbelha, TCelula, TFavo, TColmeia

// nmero mximo de clulas num favo
#define CELULAS_FAVO 3000
// dias de idade mxima de uma abelha operria
#define IDADE_MAXIMA 40
// miligramas de polen por clula (e montante carregado por uma abelha)
#define POLEN_CELULA 20
// miligramas de nectar por clula (e montante carregado por uma abelha)
#define NECTAR_CELULA 40
// miligramas de mel por clula 
#define MEL_CELULA 500
// consumo de polen e mel por dia (miligramas)
#define POLEN_ABELHA 4
#define MEL_ABELHA 10
#define TEMPO_DIA 1440

///////////////////////////////////////////////////////////////////////////////////////////////
// TAbelha: tipo que guarda informao sobre cada abelha 

// enum funcaoAbelha: tipo enumerado com a lista de funes das abelhas, por ordem do enunciado
typedef enum {
	faxineira = 0, nutriz, construtora, guardia, forrageira,
	zangao, rainha, ovo, larva, pupa, semFuncao
} TFuncaoAbelha;

typedef struct SAbelha {
	int nascimento;		
	int ultimaRefeicao; 
	TFuncaoAbelha funcao;
} TAbelha;

// NovaAbelha: melhora a legibilidade do cdigo, no seria essencial
TAbelha NovaAbelha(TFuncaoAbelha funcao, int nascimento, int ultimaRefeicao) {
	TAbelha abelha;
	abelha.funcao = funcao;
	abelha.nascimento = nascimento;
	abelha.ultimaRefeicao = ultimaRefeicao;
	return abelha;
}

// ContarAbelhas: conta o nmero de abelhas num vetor, que tem uma dada funo
int ContarAbelhas(TFuncaoAbelha funcao, TAbelha* abelhas, int tamanho) {
	int contagem = 0, i;
	for (i = 0; i < tamanho; i++)
		if (abelhas[i].funcao == funcao)
			contagem++;
	return contagem;
}

// AntigaAbelha: retorna o ndice da abelha mais antiga do vetor, numa dada funo
int AntigaAbelha(TFuncaoAbelha funcao, TAbelha* abelhas, int nAbelhas) {
	int i, abelha = -1;
	for (i = nAbelhas-1; i >= 0; i--)
		if (abelhas[i].funcao == funcao &&
			(abelha == -1 || abelhas[abelha].nascimento > abelhas[i].nascimento))
			abelha = i;
	return abelha;
}

// CriarAbelhas: cria um novo vetor de abelhas, de acordo com o enunciado
TAbelha* CriarAbelhas(int nAbelhas, int *tamanho) {
	TAbelha* abelhas = NULL;
	int i, j, id=0;
	if (nAbelhas < IDADE_MAXIMA*5) // tem de existir abelhas de todos os tipos
		return NULL;

	// tamanho do vetor
	*tamanho = 41 + (nAbelhas / IDADE_MAXIMA) * IDADE_MAXIMA;
	abelhas = (TAbelha*)malloc(sizeof(TAbelha) * (*tamanho));
	if (abelhas != NULL) {
		abelhas[id++] = NovaAbelha(rainha, 0, 0); // nascimento, ltima refeio, funo
		for (i = 1; i <= IDADE_MAXIMA; i++) { // processar todas as idades
			abelhas[id++] = NovaAbelha(zangao, -i * TEMPO_DIA, -i * TEMPO_DIA / IDADE_MAXIMA);
			for (j = 0; j < nAbelhas / IDADE_MAXIMA; j++) // h N/40 operrias em cada dia
				abelhas[id++] = NovaAbelha(
					faxineira + j % 5, // funo, vai percorrendo a lista, utilizando a ordem do tipo enumerado
					-i * TEMPO_DIA + j, // nascimento no dia certo, somando 1 minuto em cada abelha
					-i * TEMPO_DIA / IDADE_MAXIMA); // instante da ltima refeio
		}
	}
	return abelhas;
}

int InserirAbelha(TAbelha** abelhas, int* tamanho, TAbelha abelha) {
	TAbelha* novoVetor;
	int i;
	novoVetor = (TAbelha*)malloc(sizeof(TAbelha) * (*tamanho + 1));
	if (novoVetor != NULL) {
		for (i = 0; i < (*tamanho); i++)
			novoVetor[i] = (*abelhas)[i];
		novoVetor[(*tamanho)++] = abelha;
	}
	free(*abelhas);
	(*abelhas) = novoVetor;
	return (*tamanho)-1;
}

// LibertarAbelha: apaga uma abelha especfica do vetor
void LibertarAbelha(int abelha, TAbelha* abelhas, int* tamanho) {
	int i;
	if (abelha < (*tamanho)) {
		(*tamanho)--;
		for (i = abelha; i < (*tamanho); i++)
			abelhas[i] = abelhas[i + 1];
	}
}


///////////////////////////////////////////////////////////////////////////////////////////////
// TCelula: tipo que guarda informao sobre uma clula
// union: como o contedo depende do tipo de clula, optou-se pelo o union para o "conteudo"

// enum tipoCelula: tipos de clulas
typedef enum { mel = 0, polen, nectar, cria, criaZangao, realeira } TTipoCelula;

typedef struct SCelula {
	TTipoCelula tipo;
	union {
		int abelha; // ndice do ovo, larva, pupa (cria ou zango)
		int peso; // quantidade de mel, polen ou nectar, em miligramas
	} conteudo;
} TCelula;

// NovaCelula: retorna uma nova clula de um tipo
TCelula NovaCelula(TTipoCelula tipo) {
	static int quantidade[] = { MEL_CELULA, POLEN_CELULA, NECTAR_CELULA };
	TCelula celula;
	celula.tipo = tipo;
	// inicialmente vazia a clula de cria ou superior
	if (celula.tipo >= cria)
		celula.conteudo.abelha = -1;
	else 
		// clula para recursos iniciam-se cheias
		celula.conteudo.peso = quantidade[tipo];
	return celula;
}

// ContarCelulas: retorna o nmero de clulas de cada tipo e estado (vazia ou usada)
int ContarCelulas(TTipoCelula tipo, int vazia, TCelula *celulas, int tamanho) {
	int contagem = 0, estaVazio, i;
	for (i = 0; i < tamanho; i++)
		if (celulas[i].tipo == tipo) {
			estaVazio = (celulas[i].tipo >= cria ?
				celulas[i].conteudo.abelha == -1 :
				celulas[i].conteudo.peso == 0);
			if (vazia ? estaVazio : !estaVazio)
				contagem++;
		}
	return contagem;
}

// CriarCelulas: cria um vetor de clulas
TCelula* CriarCelulas(int nCelulas, int* tamanho) {
	TCelula* celulas = NULL;
	int i, j, id = 0;

	if (nCelulas <= 0) {
		(*tamanho) = 0;
		return NULL;
	}

	// tamanho do vetor
	*tamanho = 1 + nCelulas;
	celulas = (TCelula*)malloc(sizeof(TCelula) * (*tamanho));
	if (celulas != NULL) {
		celulas[id++] = NovaCelula(criaZangao); // uma clula para zanges
		for (i = 0; i < nCelulas; i++) // o resto para dividir por igual
			celulas[id++] = NovaCelula(mel + i % 4);
	}
	return celulas;
}

///////////////////////////////////////////////////////////////////////////////////////////////
// TFavo: tipo que guarda informao sobre um favo
typedef struct SFavo {
	int criacao; // data de criao
	TCelula* celulas;
	int tamanho; // nmero de clulas
} TFavo;

// NovoFavo: cria um novo favo e clulas respetivas
TFavo NovoFavo(int nCelulas, int criacao) {
	TFavo favo;
	favo.celulas = CriarCelulas(nCelulas, &favo.tamanho);
	favo.criacao = criacao;
	return favo;
}

// CriarFavos: cria um conjunto de favos, a infraestrutura da colmeia
TFavo* CriarFavos(int nCelulas, int* tamanho) {
	TFavo* favos = NULL;
	int i, id = 0, nFavos;

	if (nCelulas <= 0) 
		nFavos = (*tamanho) = 1;
	else
		nFavos = (*tamanho) = 1 + (nCelulas - 1) / CELULAS_FAVO;
	favos = (TFavo*)malloc(sizeof(TFavo) * (*tamanho));
	if (favos != NULL)
		for (i = 0; i < nFavos; i++)
			favos[id++] = NovoFavo(nCelulas / nFavos, i);
	else
		(*tamanho) = 0;
	return favos;
}

// AtualizarIDsAbelhas: todos os IDs iguais ou superior a abelha, decrementar
void AtualizarIDsAbelhas(TFavo* favos, int tamanho, int abelha) {
	int i, j;
	for (i = 0; i < tamanho; i++)
		for (j = 0; j < favos[i].tamanho; j++)
			if (favos[i].celulas[j].tipo == cria && favos[i].celulas[j].conteudo.abelha >= abelha)
				favos[i].celulas[j].conteudo.abelha--;
}


// LibertarFavos: liberta todas as clulas dos favos 
void LibertarFavos(TFavo* favos, int tamanho) {
	int i;
	for (i = 0; i < tamanho; i++)
		free(favos[i].celulas);
}


///////////////////////////////////////////////////////////////////////////////////////////////
// TColmeia: tipo que guarda informao sobre toda a colmeia
typedef struct {
	int tempo; // tempo em minutos, sendo 0 o instante de criao
	// gesto da colmeia
	int naoTrabalham; // 1 apenas h a alimentao, mas no h promoo nem trabalho
	double racioFuncao[5]; // percentagem de operrias por funo
	int prontaTrabalhar[5]; // operrias prontas para trabalhar (acumula com os minutos anteriores)
	TFavo* favos; // infraestrutura da colmeia
	int nFavos;
	TAbelha* abelhas; // recursos "humanos/abelhas" da colmeia
	int nAbelhas;
} TColmeia;

// CriarColmeia: criao de uma colmeia, com nmero de abelhas e clulas
TColmeia CriarColmeia(int nAbelhas, int nCelulas) {
	static double racioFuncao[] = { 0.1, 0.3, 0.2, 0.1, 0.3 }; // inicializar o rcios fixos
	TColmeia colmeia;
	int i;
	colmeia.tempo = 0;
	colmeia.naoTrabalham = 0;
	for (i = 0; i < 5; i++) {
		colmeia.racioFuncao[i] = racioFuncao[i];
		colmeia.prontaTrabalhar[i] = 0;
	}
	colmeia.abelhas = CriarAbelhas(nAbelhas, &colmeia.nAbelhas);
	colmeia.favos = CriarFavos(nCelulas, &colmeia.nFavos);
	return colmeia;
}

// CelulaColmeia: 
// Caso seja uma clula de recursos e abelha!=-1, retornar a clula em consumo, ou seja, 
// a clula com recursos, mas com menos recursos 
// Caso abelha seja -1 retornar uma clula desse tipo vazia
// Se for uma clula de cria, e abelha for -1 retornar uma clula vazia 
// Se abelha!=-1, procurar pela clula com essa abelha para limpar o contedo da clula
TCelula* CelulaColmeia(TTipoCelula tipo, int abelha, TColmeia* colmeia) {
	TCelula* emConsumo = NULL, * celula;
	TFavo* favo = colmeia->favos;
	int i, j;
	// ordem do mais novo para o mais velho (no h favos criados, pelo que vale a ordem de criao)
	for (i = colmeia->nFavos-1; i >= 0 ; i--) { 
		celula = favo[i].celulas;
		for (j = 0; j < favo[i].tamanho; j++) {
			if (tipo == cria) {
				if (abelha != -1) {
					if (celula[j].tipo == cria && celula[j].conteudo.abelha==abelha) {
						celula[j].conteudo.abelha = -1;
						return NULL;
					}
				}
				else if (celula[j].tipo == cria && celula[j].conteudo.abelha == -1)
					return &(celula[j]);
			}
			else if (abelha == -1) { // apenas uma clula vazia
				if (celula[j].tipo == tipo &&
					celula[j].conteudo.peso == 0)
					return &(celula[j]);
			}
			else if (celula[j].tipo == tipo &&
				celula[j].conteudo.peso > 0 &&
				(emConsumo == NULL ||
					emConsumo->conteudo.peso > celula[j].conteudo.peso))
				emConsumo = &(celula[j]);
		}
	}
	return emConsumo;
}

// AlimentarColmeia: ver que abelhas precisam de refeio, e se tem excesso de idade
void AlimentarColmeia(TColmeia* colmeia) {
	int i;
	TAbelha* abelha = colmeia->abelhas;
	TCelula* celMel = NULL, * celPolen = NULL;

	// ordem inversa de criao, mais novas so alimentadas primeiro
	for (i = colmeia->nAbelhas - 1; i >= 0; i--) {
		// cada abelha precisa de Polen e Mel por dia
		if (abelha[i].funcao > rainha) {
			// ovo, larva ou pupa, no  adulta, tem alimentao distinta 
			// (a alimentao da larva no  contemplada no e-flio)
		}
		else if (colmeia->tempo - abelha[i].nascimento >= TEMPO_DIA * IDADE_MAXIMA &&
			abelha[i].funcao != rainha) { // a rainha vive 1 a 2 anos
			// abelha morre de idade
			LibertarAbelha(i, colmeia->abelhas, &(colmeia->nAbelhas));
			AtualizarIDsAbelhas(colmeia->favos, colmeia->nFavos, i);
		}
		else if (colmeia->tempo - abelha[i].ultimaRefeicao >= TEMPO_DIA) {
			// alimentar abelha ou morre de fome
			if (celMel == NULL || celMel->conteudo.peso == 0)
				celMel = CelulaColmeia(mel, i, colmeia);
			if (celPolen == NULL || celPolen->conteudo.peso == 0)
				celPolen = CelulaColmeia(polen, i, colmeia);
			if (celMel != NULL && celPolen != NULL) {
				celMel->conteudo.peso -= MEL_ABELHA;
				celPolen->conteudo.peso -= POLEN_ABELHA;
				abelha[i].ultimaRefeicao = colmeia->tempo;
			}
			else {
				// abelha morre de fome
				LibertarAbelha(i, colmeia->abelhas, &(colmeia->nAbelhas));
				AtualizarIDsAbelhas(colmeia->favos, colmeia->nFavos, i);
			}
		}
	}
}

// PromoverColmeia: ver se h abelhas para promover
void PromoverColmeia(TColmeia* colmeia) {
	static int tempoPromocao[] = { 3,9,21 };
	int abelha, i;
	TAbelha* abelhas;
	TFuncaoAbelha funcao = forrageira;
	int nAbelhas[5]; // nmero de abelhas de cada funo (operrias)
	int totalAbelhas = 0;
	// contar abelhas de cada funo
	for (funcao = faxineira; funcao <= forrageira; funcao++) {
		nAbelhas[funcao] = ContarAbelhas(funcao, colmeia->abelhas, colmeia->nAbelhas);
		totalAbelhas += nAbelhas[funcao];
	}
	// promover se tiver a baixo do valor, promover uma abelha da funo inferior
	for (funcao = forrageira; funcao > faxineira; funcao--) 
		if (nAbelhas[funcao] < colmeia->racioFuncao[funcao] * totalAbelhas &&
			nAbelhas[funcao - 1]>0) {
			abelha = AntigaAbelha(funcao - 1, colmeia->abelhas, colmeia->nAbelhas);
			if(abelha>=0)
				colmeia->abelhas[abelha].funcao++;
		}

	// promover as abelhas no adultas
	abelhas = colmeia->abelhas;
	for (i = 0; i < colmeia->nAbelhas; i++) {
		if (abelhas[i].funcao >= ovo && abelhas[i].funcao <= pupa) {
			if (colmeia->tempo - abelhas[i].nascimento >= tempoPromocao[abelhas[i].funcao - ovo] * TEMPO_DIA)
				abelhas[i].funcao++; // passa de ovo > larva > pupa > semFuncao
			if (abelhas[i].funcao == semFuncao) { // abelha ficou adulta
				// limpar a clula onde a abelha estava
				CelulaColmeia(cria, i, colmeia);
				abelhas[i].funcao = faxineira; // passa a operria, primeiro escalo
				abelhas[i].nascimento = colmeia->tempo; // nascimento adulto
			}
		}

	}
}

// PorOvosColmeia: colocao de um ovo numa clula de cria
void PorOvosColmeia(TColmeia* colmeia) {
	// necessrio 10 faxineiras para limpar, e nesse caso a rainha coloca um ovo
	TCelula* celula = CelulaColmeia(cria, -1, colmeia);
	if (celula != NULL && colmeia->prontaTrabalhar[faxineira] >= 10 &&
		ContarAbelhas(rainha, colmeia->abelhas, colmeia->nAbelhas) > 0) { // tem de ainda existir a rainha
		colmeia->prontaTrabalhar[faxineira] -= 10;
		celula->conteudo.abelha = InserirAbelha(
			&(colmeia->abelhas), 
			&(colmeia->nAbelhas),
			NovaAbelha(ovo, colmeia->tempo, colmeia->tempo));
	}
}

// ProcurarRecursosColmeia: procurar por polen ou nectar
void ProcurarRecursosColmeia(TColmeia *colmeia) {
	TCelula* celula;
	int procNectar = 0; // procura sempre por plen, 
	// necessrio 60 minutos para uma abelha ir buscar polen ou nectar, considerar 60 abelhas.minuto
	// a clula tem tambm de ser limpa
	celula = CelulaColmeia(procNectar ? nectar : polen, -1, colmeia);
	if (celula != NULL && colmeia->prontaTrabalhar[forrageira] >= 60 ) {
		colmeia->prontaTrabalhar[forrageira] -= 60;
		colmeia->prontaTrabalhar[faxineira] -= 10; // se faxineiras no estiverem disponveis, a rainha no coloca ovos
		celula->conteudo.peso = (procNectar ? NECTAR_CELULA : POLEN_CELULA);
	}
	else if (celula == NULL)
		colmeia->prontaTrabalhar[forrageira] = 0; // quantidade de trabalho desperdiada
}


// TrabalharColmeia: atualizar quantidade de trabalho de cada tipo disponvel, e 
// - rainha pode num minuto, colocar um novo ovo, se existir uma clula de cria ou criaZangao limpa
// - nutrizes, alimentar as larvas (do dia 3 ao dia 9), passando depois a pulpa --- cada alimentao 1 minuto,
//   tendo de passar 15 vezes, em cada vez o consumo dirio de uma abelha adulta --- larvas no alimentadas, morrem
// - construtoras, se existirem 10 ou mais livres, iniciar construo de uma nova clula, no favo mais vazio
//   procurar respeitar a precentagem ideal de tipos de clulas. 
// - guardians, se existir perigo, atacar
// - forrageiras, procurar nectar/polen na proporo ideal (depende das que esto em transito)
// - em cada viagem, o tempo de viagem  entre 30 e 60 minutos
void TrabalharColmeia(TColmeia* colmeia) {
	TAbelha* abelha = colmeia->abelhas;
	TFuncaoAbelha funcao;
	int i;
	for (i = 0; i < colmeia->nAbelhas; i++)
		if (abelha[i].funcao <= forrageira && abelha[i].ultimaRefeicao != colmeia->tempo)
			colmeia->prontaTrabalhar[abelha[i].funcao]++;
	// operaes
	PorOvosColmeia(colmeia);
	//	AlimentarLarvasColmeia(colmeia); // (nut) fora do e-flio 
	//	ConstruirColmeia(colmeia); // (con) fora do e-flio
	ProcurarRecursosColmeia(colmeia);
	// .... (gua), nectar>mel...
}

// LibertarColmeia: liberta todas as estruturas de dados
void LibertarColmeia(TColmeia* colmeia) {
	LibertarFavos(colmeia->favos, colmeia->nFavos);
	free(colmeia->abelhas);
}

///////////////////////////////////////////////////////////////////////////
// Relatrio

// MostraTempo: funo para mostrar os minutos agregados em horas e dias
void MostraTempo(int minutos) {
	int dias, horas;
	dias = minutos / TEMPO_DIA;
	minutos -= dias * TEMPO_DIA;
	horas = minutos / 60;
	minutos -= horas * 60;
	if (dias > 0)
		printf("%d dia(s) ", dias);
	if (horas > 0)
		printf("%d hora(s) ", horas);
	if (minutos > 0)
		printf("%d minuto(s)", minutos);
}

void MostrarQuantidade(int valor) {
	if (valor > 0)
		printf(" %3d", valor);
	else
		printf("    ");
}

// RelatorioCelulas: mostra a parte do relatrio de favos, das suas clulas
void RelatorioCelulas(TCelula* celulas, int tamanho) {
	printf("            celulas vazias:");
	for (TTipoCelula tipo = mel; tipo <= realeira; tipo++)
		MostrarQuantidade(ContarCelulas(tipo, 1, celulas, tamanho));
	printf("\n                                 usadas:");
	for (TTipoCelula tipo = mel; tipo <= realeira; tipo++)
		MostrarQuantidade(ContarCelulas(tipo, 0, celulas, tamanho));
}

// RelatorioFavos: mostra o relatrio do favo com determinada data de criao
void RelatorioFavos(TFavo* favos, int tamanho, int criacao) {
	int i;
	for (i = 0; i < tamanho; i++)
		if (favos[i].criacao == criacao) {
			printf("\n    Favo %3d:", favos[i].criacao);
			RelatorioCelulas(favos[i].celulas, favos[i].tamanho);
		}
}

// RelatorioColmeia: mostra o relatrio da colmeia, o nico solicitado no e-flio
void RelatorioColmeia(TColmeia* colmeia) {
	int criacao;
	printf("\nRelatorio colmeia: ");
	MostraTempo(colmeia->tempo);
	printf("\n fax nut con gua for zan rai ovo lar pup mel pol nec cri zan rea \n");
	for (TFuncaoAbelha funcao = faxineira; funcao < semFuncao; funcao++)
		MostrarQuantidade(ContarAbelhas(funcao, colmeia->abelhas, colmeia->nAbelhas));
	for (criacao = 0; criacao < colmeia->nFavos; criacao++)
		RelatorioFavos(colmeia->favos, colmeia->nFavos, criacao);
}

///////////////////////////////////////////////////////////////////////////
// Simulao

// SimularColmeia: avana 1 minuto simulando as operaes na colmeia
void SimularColmeia(TColmeia* colmeia) {
	colmeia->tempo++; // avana o tempo

	// - alimentar todas as abelhas que esteja na colmeia, se j passou 24h desde a ultima refeio
	// - caso uma abelha esteja em excesso de idade, remover em vez de alimentar
	AlimentarColmeia(colmeia);

	if (colmeia->naoTrabalham == 0) { // alnea D
		// - promover abelhas por idade, procurando respeitar a proporo pretendida entre funes
		PromoverColmeia(colmeia);

		// coloca as abelhas a trabalhar, conforme a funo
		TrabalharColmeia(colmeia);
	}
}

// TemporadaColmeia: simula uma temporada, parando caso no existam abelhas
void TemporadaColmeia(int tempo, TColmeia* colmeia) {
	while (tempo > 0 && colmeia->nAbelhas > 0) {
		tempo--;
		SimularColmeia(colmeia);
	}
}

// FilmeColmeia: sequncia de temporadas intercaladas com um relatrio de colmeia
void FilmeColmeia(int tempo, int temporadas, TColmeia* colmeia) {
	RelatorioColmeia(colmeia);
	while (--temporadas > 0 && colmeia->nAbelhas > 0) {
		TemporadaColmeia(tempo, colmeia);
		RelatorioColmeia(colmeia);
	}
}

///////////////////////////////////////////////////////////////////////////
// main: execuo das 4 alneas
int main() 
{
	int nAbelhas, nCelulas, minutos, temporadas;
	TColmeia colmeia;

	//printf("Quantidade de abelhas: ");
	scanf("%d", &nAbelhas);
	//printf("Quantidade de celulas: ");
	scanf("%d", &nCelulas);
	//printf("Minutos por temporada: ");
	scanf("%d", &minutos);
	//printf("Temporadas: ");
	scanf("%d", &temporadas);
	colmeia = CriarColmeia(nAbelhas, nCelulas);
	if (temporadas < 0) { // alnea C
		temporadas *= -1;
		colmeia.naoTrabalham = 1;
	}
	FilmeColmeia(minutos, temporadas + 1, &colmeia);
	LibertarColmeia(&colmeia);
}
